# **************************************************************************
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************
import numpy as np

from continuousflex.protocols.protocol_genesis import *
import pyworkflow.protocol.params as params
from sklearn import decomposition
from xmipp3.convert import writeSetOfVolumes, writeSetOfParticles, readSetOfVolumes, readSetOfParticles
from pwem.constants import ALIGN_PROJ
from continuousflex.protocols.convert import matrix2eulerAngles
from pwem.emlib import MetaData, MDL_ENABLED, MDL_NMA_MODEFILE,MDL_ORDER
from pwem.objects import SetOfNormalModes
from .convert import rowToMode
from xmipp3.base import XmippMdRow

class FlexProtMDSPACE(FlexProtGenesis):

    """ Protocol to perform MDSPACE using GENESIS """
    _label = 'MDSPACE'

    def __init__(self, **kwargs):
        FlexProtGenesis.__init__(self, **kwargs)
        self._iter = 0
        self._missing_pdbs = None

        # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='MDSPACE Refinement')

        form.addParam('numberOfIter', params.IntParam, label="Number of iterations", default=3,
                      help="Number of round of fitting for MDSPACE", important=True)

        form.addParam('numberOfPCA', params.IntParam, label="Number of PCA component", default=5,
                      help="Number of principal component to keep at each round", important=True)

        FlexProtGenesis._defineParams(self, form)


    def _insertAllSteps(self):
        # make path
        self._insertFunctionStep("makePathStep")

        # Convert input PDB
        self._insertFunctionStep("convertInputPDBStep")

        # Convert normal modes
        if (self.simulationType.get() == SIMULATION_NMMD or self.simulationType.get() == SIMULATION_RENMMD):
            self._insertFunctionStep("convertNormalModeFileStep")

        # Convert input EM data
        if self.EMfitChoice.get() != EMFIT_NONE:
            self._insertFunctionStep("convertInputEMStep")

        for iter_global in range(self.numberOfIter.get()):

            # Create INP files
            self._insertFunctionStep("createGenesisInputStep")

            # RUN simulation
            if self.parallelType.get() == PARALLEL_MPI and self.getNumberOfSimulation() > 1:
                self._insertFunctionStep("runSimulationMPI")
            elif self.parallelType.get() == PARALLEL_GNU and self.getNumberOfSimulation() > 1 and existsCommand(
                    "parallel"):
                self._insertFunctionStep("runSimulationParallel")
            else:
                for i in range(self.getNumberOfSimulation()):
                    self._insertFunctionStep("runSimulation", self.getGenesisInputFile(i), self.getOutputPrefix(i))

            self._insertFunctionStep("pdb2dcdStep")

            self._insertFunctionStep("rigidBodyAlignementStep")

            self._insertFunctionStep("updateAlignementStep")

            self._insertFunctionStep("PCAStep")

            self._insertFunctionStep("runMinimizationStep")

            self._insertFunctionStep("newIterationStep")

        self._insertFunctionStep("createOutputStep")

    def pdb2dcdStep(self):
        pdbs_matrix = []
        missing_pdbs = []

        for i in range(self.getNumberOfSimulation()):
            pdb_fname = self.getOutputPrefix(i) +".pdb"
            if os.path.isfile(pdb_fname) and os.path.getsize(pdb_fname) != 0:
                mol = ContinuousFlexPDBHandler(pdb_fname)
                pdbs_matrix.append(mol.coords)
            else:
                missing_pdbs.append(i)

        pdbs_arr = np.array(pdbs_matrix)

        # save as dcd file
        numpyArr2dcd(pdbs_arr, self._getExtraPath("coords.dcd"))

        # If some pdbs are missing (fitting failed), save indexes
        self._missing_pdbs = np.array(missing_pdbs).astype(int)
        print("MiSSING ARRAY : ")
        print(self._missing_pdbs)

    def rigidBodyAlignementStep(self):

        # open files
        refPDB =  ContinuousFlexPDBHandler(self.getInputPDBprefix()+".pdb")
        arrDCD = dcd2numpyArr(self._getExtraPath("coords.dcd"))
        nframe, natom,_ =arrDCD.shape
        alignXMD = md.MetaData()

        # loop over all pdbs
        for i in range(nframe):
            print("Aligning PDB %i ... " %i)

            # rotate
            coord = arrDCD[i]
            rot_mat, tran = ContinuousFlexPDBHandler.alignCoords(refPDB.coords, coord)
            arrDCD[i] = (np.dot(arrDCD[i], rot_mat) + tran).astype(np.float32)

            # add to MD
            trans_mat = np.zeros((4,4))
            trans_mat[:3,:3] = rot_mat
            trans_mat[:,3][:3] = tran
            rot, tilt, psi,shftx, shfty, shftz = matrix2eulerAngles(trans_mat)
            index = alignXMD.addObject()
            alignXMD.setValue(md.MDL_ANGLE_ROT, rot, index)
            alignXMD.setValue(md.MDL_ANGLE_TILT, tilt, index)
            alignXMD.setValue(md.MDL_ANGLE_PSI, psi, index)
            alignXMD.setValue(md.MDL_SHIFT_X, shftx, index)
            alignXMD.setValue(md.MDL_SHIFT_Y, shfty, index)
            alignXMD.setValue(md.MDL_SHIFT_Z, shftz, index)
            alignXMD.setValue(md.MDL_IMAGE, "", index)

        numpyArr2dcd(arrDCD, self._getExtraPath("coords.dcd"))
        alignXMD.write(self.getTransformation())

    def updateAlignementStep(self):

        if self._iter == 0:
            inputSet = self.inputImage.get()
        else:
            inputSet = self._createSetOfParticles("inputSet")
            readSetOfParticles(self.getAlignementPrefix(self._iter-1), inputSet)
            inputSet.setSamplingRate(self.inputImage.get().getSamplingRate())

        inputTransformation = self._createSetOfParticles("inputTransformation")
        readSetOfParticles(self.getTransformation(), inputTransformation)
        alignedSet = self._createSetOfParticles("alignedSet")

        alignedSet.setSamplingRate(inputSet.getSamplingRate())
        alignedSet.setAlignment(ALIGN_PROJ)
        iter1 = inputSet.iterItems()
        iter2 = inputTransformation.iterItems()
        for i in range(self.getNumberOfSimulation()):
            p1 = iter1.__next__()
            r1 = p1.getTransform()
            if not i in self._missing_pdbs :
                p2 = iter2.__next__()
                r2 = p2.getTransform()
                rot = r2.getRotationMatrix()
                tran = np.array(r2.getShifts()) / inputSet.getSamplingRate()
                new_trans = np.zeros((4, 4))
                new_trans[:3, 3] = tran
                new_trans[:3, :3] = rot
                new_trans[3, 3] = 1.0
                r1.composeTransform(new_trans)
            else:
                print("MISSING PDBS %s"%(i+1))
            p1.setTransform(r1)
            alignedSet.append(p1)

        writeSetOfParticles(alignedSet, self.getAlignementPrefix())
        self._inputEMMetadata = md.MetaData(self.getAlignementPrefix())

    def PCAStep(self):

        numberOfPCA = self.numberOfPCA.get()

        pdbs_arr = dcd2numpyArr(self._getExtraPath("coords.dcd"))
        nframe, natom,_ = pdbs_arr.shape
        pdbs_matrix = pdbs_arr.reshape(nframe, natom*3)

        pca = decomposition.PCA(n_components=numberOfPCA)
        Y = pca.fit_transform(pdbs_matrix)

        pdb = ContinuousFlexPDBHandler(self.getInputPDBprefix()+".pdb")
        pdb.coords = pca.mean_.reshape(pdbs_matrix.shape[1] // 3, 3)

        matrix = pca.components_.reshape(numberOfPCA,pdbs_matrix.shape[1]//3,3)

        # SAVE NEW inputs
        pca_prefix = self.getPCAPrefix()
        pdb.write_pdb(pca_prefix+".pdb")
        nm_file = pca_prefix+".nma"
        np.savetxt(pca_prefix+"_matrix.txt", Y)
        with open(nm_file, "w") as f:
            for i in range(numberOfPCA):
                f.write(" VECTOR    %i       VALUE  0.0\n" % (i + 1))
                f.write(" -----------------------------------\n")
                for j in range(matrix.shape[1]):
                    f.write(" %e   %e   %e\n" %  (matrix[i,j, 0], matrix[i,j, 1], matrix[i,j, 2]))

    def runMinimizationStep(self):

        # INP file name
        inp_file = "%s/INP_min"%self.getGenFiles()

        # copy inputs
        pcapref = self.getPCAPrefix()
        tmppref = self._getExtraPath("tmp")
        runCommand("cp %s.pdb %s.pdb"%(pcapref, tmppref))
        runCommand("cp %s.nma %s.nma"%(pcapref, tmppref))
        if self.getForceField() == FORCEFIELD_CHARMM:
            runCommand("cp %s.psf %s.psf"%(self.getInputPDBprefix(), tmppref))
        else:
            runCommand("cp %s.top %s.top"%(self.getInputPDBprefix(), tmppref))

        # Set Inputs files
        args = self.getDefaultArgs()
        args["inputPDBprefix"]  = tmppref
        args["outputPrefix"]  = self.getInputPDBprefix()+"_min"
        args["simulationType"]  = SIMULATION_MIN
        args["inputType"]  = INPUT_NEW_SIM
        args["n_steps"]  = 10000
        args["EMfitChoice"]  = EMFIT_NONE

        # Create input genesis file
        createGenesisInput(inp_file, **args)

        # Run minimization
        self.runSimulation(inp_file, self.getInputPDBprefix()+"_min")

    def newIterationStep(self):
        for i in range(self.getNumberOfSimulation()):
            outpref = self._getExtraPath("output_%s" % str(i + 1).zfill(6))
            outpref_itr =self.getOutputPrefix(i)
            runCommand("cat %s.log >> %s.log" % (outpref_itr,outpref))

            dcdfile_itr = outpref_itr+ ".dcd"
            if os.path.isfile(dcdfile_itr):
                dcdfile = outpref+".dcd"
                if os.path.isfile(dcdfile):
                    try:
                        dcdarr = np.concatenate((dcd2numpyArr(dcdfile),dcd2numpyArr(dcdfile_itr)), axis=0)
                        numpyArr2dcd(dcdarr, dcdfile)
                    except ValueError:
                        print("Incomplete DCD file")
                else:
                    runCommand("cp %s %s"%(dcdfile_itr, dcdfile))

            pdbfile = outpref_itr+".pdb"
            if os.path.isfile(pdbfile):
                runCommand("cp %s %s.pdb" % (pdbfile, outpref))

        inputPref = self.getInputPDBprefix()
        pcaPref = self.getPCAPrefix()

        ############# INCREMENT ITERATION ###################
        if self._iter < self.numberOfIter.get()-1:
            self._iter += 1
            print("New Iteration %i"%self._iter)
            ###################################################

            inputPref_incr = self.getInputPDBprefix()

            runCommand("cp %s_min.pdb %s.pdb"%(inputPref, inputPref_incr))
            runCommand("cp %s_min.rst %s.rst"%(inputPref, inputPref_incr))
            runCommand("cp %s.nma %s.nma"%(pcaPref, inputPref_incr))
            if self.getForceField() == FORCEFIELD_CHARMM:
                runCommand("cp %s.psf %s.psf" % (inputPref, inputPref_incr))
            elif self.getForceField() == FORCEFIELD_CAGO or self.getForceField() == FORCEFIELD_AAGO :
                runCommand("cp %s.top %s.top" % (inputPref, inputPref_incr))

    def createGenesisInputStep(self):
        """
        Create GENESIS input files
        :return None:
        """
        for indexFit in range(self.getNumberOfSimulation()):
            inp_file = self.getGenesisInputFile(indexFit)
            args = self.getDefaultArgs(indexFit)
            if self._iter != 0 :
                args["inputType"] = INPUT_NEW_SIM
                args["simulationType"] = SIMULATION_NMMD
                args["nm_number"] =  self.numberOfPCA.get()
                args["nm_dt"] =  0.002
                args["nm_mass"] =  5.0
            createGenesisInput(inp_file, **args)

    def createOutputStep(self):
        FlexProtGenesis.createOutputStep(self)

        runCommand("cp %s.pdb %s"%(self.getPCAPrefix(), self.getPath("atoms.pdb")))
        pdb = AtomStruct(self._getPath("atoms.pdb"))
        natoms = ContinuousFlexPDBHandler(self._getPath("atoms.pdb")).n_atoms
        self._defineOutputs(outputMean=pdb)

        makePath(self._getPath("modes"))
        pc_file = self.getPCAPrefix()+".nma"
        with open(pc_file, "r") as f:
            for i in range(self.numberOfPCA.get()):
                f.readline()
                f.readline()
                modefile = self._getPath("modes", "vec.%d" % (i + 1))
                with open(modefile, "w") as fout:
                    for j in range(natoms):
                        fout.write(f.readline())
        mdOut = MetaData()
        for i in range(self.numberOfPCA.get()):
            objId = mdOut.addObject()
            modefile = self._getPath("modes", "vec.%d" % (i + 1))
            mdOut.setValue(MDL_NMA_MODEFILE, modefile, objId)
            mdOut.setValue(MDL_ORDER, i + 1, objId)
            mdOut.setValue(MDL_ENABLED, 1, objId)
        mdOut.write(self._getPath("modes.xmd"))
        pcSet = SetOfNormalModes(filename=self._getPath("modes.sqlite"))
        row = XmippMdRow()
        for objId in mdOut:
            row.readFromMd(mdOut, objId)
            pcSet.append(rowToMode(row))
        pcSet.setPdb(pdb)
        self._defineOutputs(outputPCA=pcSet)

    def getOutputPrefix(self, index=0):
        return self._getExtraPath("output_%s_iter_%s" % (str(index + 1).zfill(6),str(self._iter+1).zfill(3)))

    def getInputPDBprefix(self, index=0):
        """
        Get the input PDB prefix of the specified index
        :param int index: index of input PDB
        :return str: Input PDB prefix
        """
        prefix = self._getExtraPath("inputPDB_%s_iter_%s")
        if self.getNumberOfInputPDB() == 1:
            return prefix % (str(1).zfill(6),str(self._iter+1).zfill(3))
        else:
            return prefix % (str(index + 1).zfill(6),str(self._iter+1).zfill(3))
    def getPCAPrefix(self):
        return self._getExtraPath("pca_iter_%s" % (str(self._iter+1).zfill(3)))

    def getAlignementPrefix(self, itr=None):
        if itr is None : itr = self._iter
        return "%s/alignement_iter_%s.xmd"%(self.getEmdFiles(),str(itr+1).zfill(3))
    def getTransformation(self, itr=None):
        if itr is None : itr = self._iter
        return "%s/transformation_iter_%s.xmd"%(self.getEmdFiles(),str(itr+1).zfill(3))

    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _validate(self):
        errors = []
        return errors

    def _citations(self):
        return ['vuillemot2023mdspace','vuillemot2022NMMD','harastani2022continuousflex']

    def _methods(self):
        return []


# **************************************************************************
# * Authors:     Rémi Vuillemot (remi.vuillemot@upmc.fr)
# * IMPMC, Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************

from pwem.protocols import ProtImportPdb
from pwem.tests.workflows import TestWorkflow
from pyworkflow.tests import setupTestProject, DataSet

from continuousflex.protocols.protocol_mdspace import FlexProtMDSPACE
from continuousflex.protocols import FlexProtNMA, NMA_CUTOFF_ABS, FlexProtSynthesizeImages, \
    FlexProtDimredPdb, FlexProtAlignPdb,FlexProtGenesis
from continuousflex.protocols.utilities.genesis_utilities import *

from continuousflex.protocols.protocol_align_pdbs import PDB_SOURCE_OBJECT
from continuousflex.protocols.protocol_pdb_dimred import PDB_SOURCE_ALIGNED, REDUCE_METHOD_PCA, REDUCE_METHOD_UMAP

class TestMDSPACE(TestWorkflow):
    """ Test Class for MDSPACE. """

    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        cls.        ds = DataSet.getDataSet('nma_V2.0')


    def test_MDSPACE(self):


        # ------------------------- Import PDB prot --------------------------------
        protPdb4ake = self.newProtocol(ProtImportPdb, inputPdbData=1,
                                      pdbFile=self.ds.getFile('4ake_ca_pdb'))
        protPdb4ake.setObjLabel('Input PDB (4AKE C-Alpha only)')
        self.launchProtocol(protPdb4ake)
        # ------------------------- Genesis Min prot --------------------------------

        protGenesisMin = self.newProtocol(FlexProtGenesis,
                                          inputPDB=protPdb4ake.outputPdb,
                                          forcefield=FORCEFIELD_CAGO,
                                          inputType=INPUT_NEW_SIM,
                                          inputTOP=self.ds.getFile('4ake_ca_top'),

                                          simulationType=SIMULATION_MIN,
                                          time_step=0.001,
                                          n_steps=100,
                                          eneout_period=10,
                                          crdout_period=10,
                                          nbupdate_period=10,

                                          implicitSolvent=IMPLICIT_SOLVENT_NONE,
                                          electrostatics=ELECTROSTATICS_CUTOFF,
                                          switch_dist=10.0,
                                          cutoff_dist=12.0,
                                          pairlist_dist=15.0,

                                          numberOfThreads=NUMBER_OF_CPU,
                                          numberOfMpi=1,
                                          )
        protGenesisMin.setObjLabel('Energy Minimization CAGO')
        # Launch minimisation
        self.launchProtocol(protGenesisMin)

        # ------------------------- NMA prot --------------------------------
        # Launch NMA for energy min PDB
        protNMA = self.newProtocol(FlexProtNMA,
                                   cutoffMode=NMA_CUTOFF_ABS)
        protNMA.inputStructure.set(protGenesisMin.outputPDB)
        protNMA.setObjLabel('NMA')
        self.launchProtocol(protNMA)

        # ------------------------- synth images --------------------------------
        target_images = self.newProtocol(FlexProtSynthesizeImages,
                                         inputModes=protNMA.outputModes,
                                         numberOfVolumes=100,
                                         samplingRate=2.0,
                                         modesAmplitudeRange=50,
                                         seedOption=False,
                                         noiseCTFChoice=1,
                                         volumeSize=64)
        target_images.setObjLabel('Target particles')
        self.launchProtocol(target_images)

        # ------------------------- MDSPACE --------------------------------
        protMDSPACE = self.newProtocol(FlexProtMDSPACE,

                                                 inputType=INPUT_RESTART,
                                                 restartProt=protGenesisMin,

                                                 simulationType=SIMULATION_NMMD,
                                                 time_step=0.001,
                                                 n_steps=5000,
                                                 eneout_period=100,
                                                 crdout_period=100,
                                                 nbupdate_period=10,
                                                 nm_number=6,
                                                 nm_mass=1.0,
                                                 inputModes=protNMA.outputModes,

                                                 implicitSolvent=IMPLICIT_SOLVENT_NONE,
                                                 electrostatics=ELECTROSTATICS_CUTOFF,
                                                 switch_dist=10.0,
                                                 cutoff_dist=12.0,
                                                 pairlist_dist=15.0,

                                                 ensemble=ENSEMBLE_NVT,
                                                 tpcontrol=TPCONTROL_LANGEVIN,
                                                 temperature=50.0,

                                                 boundary=BOUNDARY_NOBC,
                                                 EMfitChoice=EMFIT_IMAGES,
                                                 constantK="100",
                                                 emfit_sigma=2.0,
                                                 emfit_tolerance=0.1,
                                                 inputImage=target_images.outputImages,
                                                 pixel_size=2.0,
                                                 imageAngleShift=target_images._getExtraPath("GroundTruth.xmd"),

                                                 numberOfThreads=1,
                                                 numberOfMpi=NUMBER_OF_CPU,
                                                 )
        protMDSPACE.setObjLabel('MDSPACE')

        # Launch Fitting
        self.launchProtocol(protMDSPACE)

        # ------------------------- align pdbs --------------------------------
        alignPDBs = self.newProtocol(FlexProtAlignPdb,
                                     pdbSource = PDB_SOURCE_OBJECT,
                                     setOfPDBs = protMDSPACE.outputPDBs,
                                     alignRefPDB=protGenesisMin.outputPDB,
                                     createOutput=False)
        alignPDBs.setObjLabel('Align output PDBs')
        self.launchProtocol(alignPDBs)

        # ------------------------- PCA --------------------------------
        protPca = self.newProtocol(FlexProtDimredPdb,
                                   pdbSource=PDB_SOURCE_ALIGNED,
                                   alignPdbProt=alignPDBs,
                                   method=REDUCE_METHOD_PCA)
        protPca.setObjLabel('PCA')
        self.launchProtocol(protPca)
        # ------------------------- UMAP --------------------------------
        protUmap = self.newProtocol(FlexProtDimredPdb,
                                   pdbSource=PDB_SOURCE_ALIGNED,
                                   alignPdbProt=alignPDBs,
                                   method=REDUCE_METHOD_UMAP)
        protUmap.setObjLabel('UMAP')
        self.launchProtocol(protUmap)


# By Remi Vuillemot

import numpy as np
from pyworkflow.utils import runCommand
import multiprocessing

NUMBER_OF_CPU = int(np.min([multiprocessing.cpu_count(),4]))

EMFIT_NONE = 0
EMFIT_VOLUMES = 1
EMFIT_IMAGES = 2

FORCEFIELD_CHARMM = 0
FORCEFIELD_AAGO = 1
FORCEFIELD_CAGO = 2

SIMULATION_MIN = 0
SIMULATION_MD = 1
SIMULATION_NMMD = 2
SIMULATION_REMD = 3
SIMULATION_RENMMD = 4

PROGRAM_ATDYN = 0
PROGRAM_SPDYN= 1

INTEGRATOR_VVERLET = 0
INTEGRATOR_LEAPFROG = 1

IMPLICIT_SOLVENT_GBSA = 0
IMPLICIT_SOLVENT_NONE = 1

TPCONTROL_NONE = 0
TPCONTROL_LANGEVIN = 1
TPCONTROL_BERENDSEN = 2
TPCONTROL_BUSSI = 3

ENSEMBLE_NVT = 0
ENSEMBLE_NVE = 1
ENSEMBLE_NPT = 2

BOUNDARY_NOBC = 0
BOUNDARY_PBC = 1

ELECTROSTATICS_PME = 0
ELECTROSTATICS_CUTOFF = 1

NUCLEIC_NO = 0
NUCLEIC_RNA =1
NUCLEIC_DNA = 2

RB_PROJMATCH = 0
RB_WAVELET = 1

INPUT_TOPOLOGY = 0
INPUT_RESTART = 1
INPUT_NEW_SIM = 2

PROJECTION_ANGLE_SAME=0
PROJECTION_ANGLE_XMIPP=1
PROJECTION_ANGLE_IMAGE=2

PARALLEL_MPI = 0
PARALLEL_GNU = 1
PARALLEL_SERIAL = 2


def lastPDBFromDCD(inputPDB,inputDCD,  outputPDB):

    # EXTRACT PDB from dcd file
    with open("%s_tmp_dcd2pdb.tcl" % outputPDB, "w") as f:
        s = ""
        s += "mol load pdb %s dcd %s\n" % (inputPDB, inputDCD)
        s += "set nf [molinfo top get numframes]\n"
        s += "[atomselect top all frame [expr $nf - 1]] writepdb %s\n" % outputPDB
        s += "exit\n"
        f.write(s)
    runCommand("vmd -dispdev text -e %s_tmp_dcd2pdb.tcl" % outputPDB)

    # CLEAN TMP FILES
    runCommand("rm -f %s_tmp_dcd2pdb.tcl" % (outputPDB))


def readLogFile(log_file):
    with open(log_file,"r") as file:
        header = None
        dic = {}
        for line in file:
            if line.startswith("INFO:"):
                if header is None:
                    header = line.split()
                    for i in range(1,len(header)):
                        dic[header[i]] = []
                else:
                    splitline = line.split()
                    if len(splitline) >= len(header):
                        for i in range(1,len(header)):
                            try :
                                dic[header[i]].append(float(splitline[i]))
                            except ValueError:
                                pass
    return dic
def dcd2numpyArr(filename):
    print("> Reading dcd file %s"%filename)
    BYTESIZE = 4
    with open(filename, 'rb') as f:

        # Header
        # ---------------- INIT

        start_size = int.from_bytes((f.read(BYTESIZE)), "little")
        crd_type = f.read(BYTESIZE).decode('ascii')
        nframe = int.from_bytes((f.read(BYTESIZE)), "little")
        start_frame = int.from_bytes((f.read(BYTESIZE)), "little")
        len_frame = int.from_bytes((f.read(BYTESIZE)), "little")
        len_total = int.from_bytes((f.read(BYTESIZE)), "little")
        for i in range(5):
            f.read(BYTESIZE)
        time_step = np.frombuffer(f.read(BYTESIZE), dtype=np.float32)
        for i in range(9):
            f.read(BYTESIZE)
        charmm_version = int.from_bytes((f.read(BYTESIZE)), "little")

        end_size = int.from_bytes((f.read(BYTESIZE)), "little")

        if end_size != start_size:
            raise RuntimeError("Can not read dcd file")

        # ---------------- TITLE
        start_size = int.from_bytes((f.read(BYTESIZE)), "little")
        ntitle = int.from_bytes((f.read(BYTESIZE)), "little")
        tilte_rd = f.read(BYTESIZE*20 * ntitle)
        try :
            title = tilte_rd.encode("ascii")
        except AttributeError:
            title = str(tilte_rd)
        end_size = int.from_bytes((f.read(BYTESIZE)), "little")

        if end_size != start_size:
            raise RuntimeError("Can not read dcd file")

        # ---------------- NATOM
        start_size = int.from_bytes((f.read(BYTESIZE)), "little")
        natom = int.from_bytes((f.read(BYTESIZE)), "little")
        end_size = int.from_bytes((f.read(BYTESIZE)), "little")

        if end_size != start_size:
            raise RuntimeError("Can not read dcd file")

        # ----------------- DCD COORD
        dcd_arr =  np.zeros((nframe, natom, 3), dtype=np.float32)
        for i in range(nframe):
            for j in range(3):

                start_size = int.from_bytes((f.read(BYTESIZE)), "little")
                while (start_size != BYTESIZE * natom and start_size != 0):
                    # print("\n-- UNKNOWN %s -- " % start_size)

                    f.read(start_size)
                    end_size = int.from_bytes((f.read(BYTESIZE)), "little")
                    if end_size != start_size:
                        raise RuntimeError("Can not read dcd file")
                    start_size = int.from_bytes((f.read(BYTESIZE)), "little")

                bin_arr = f.read(BYTESIZE * natom)
                if len(bin_arr) == BYTESIZE * natom:
                    dcd_arr[i, :, j] = np.frombuffer(bin_arr, dtype=np.float32)
                else:
                    break
                end_size = int.from_bytes((f.read(BYTESIZE)), "little")
                if end_size != start_size:
                    if i>1:
                        break
                    else:
                        # pass
                        raise RuntimeError("Can not read dcd file %i %i " % (start_size, end_size))

        print("\t -- Summary of DCD file -- ")
        print("\t\t crd_type  : %s"%crd_type)
        print("\t\t nframe  : %s"%nframe)
        print("\t\t len_frame  : %s"%len_frame)
        print("\t\t len_total  : %s"%len_total)
        print("\t\t time_step  : %s"%time_step)
        print("\t\t charmm_version  : %s"%charmm_version)
        print("\t\t title  : %s"%title)
        print("\t\t natom  : %s"%natom)
    print("\t Done \n")

    return dcd_arr

def numpyArr2dcd(arr, filename, start_frame=1, len_frame=1, time_step=1.0, title=None):
    print("> Wrinting dcd file %s"%filename)
    BYTESIZE = 4
    nframe, natom, _ = arr.shape
    len_total=nframe*len_frame
    charmm_version=24
    if title is None:
        title = "DCD file generated by Continuous Flex plugin"
    ntitle = (len(title)//(20*BYTESIZE)) + 1
    with open(filename, 'wb') as f:
        zeroByte = int.to_bytes(0, BYTESIZE, "little")

        # Header
        # ---------------- INIT
        f.write(int.to_bytes(21*BYTESIZE ,BYTESIZE, "little"))
        f.write(b'CORD')
        f.write(int.to_bytes(nframe, BYTESIZE, "little"))
        f.write(int.to_bytes(start_frame, BYTESIZE, "little"))
        f.write(int.to_bytes(len_frame, BYTESIZE, "little"))
        f.write(int.to_bytes(len_total, BYTESIZE, "little"))
        for i in range(5):
            f.write(zeroByte)
        f.write(np.float32(time_step).tobytes())
        for i in range(9):
            f.write(zeroByte)
        f.write(int.to_bytes(charmm_version, BYTESIZE, "little"))

        f.write(int.to_bytes(21*BYTESIZE,BYTESIZE, "little"))

        # ---------------- TITLE
        f.write(int.to_bytes((ntitle*20+1)*BYTESIZE ,BYTESIZE, "little"))
        f.write(int.to_bytes(ntitle ,BYTESIZE, "little"))
        f.write(title.ljust(20*BYTESIZE).encode("ascii"))
        f.write(int.to_bytes((ntitle*20+1)*BYTESIZE ,BYTESIZE, "little"))

        # ---------------- NATOM
        f.write(int.to_bytes(BYTESIZE ,BYTESIZE, "little"))
        f.write(int.to_bytes(natom ,BYTESIZE, "little"))
        f.write(int.to_bytes(BYTESIZE ,BYTESIZE, "little"))

        # ----------------- DCD COORD
        for i in range(nframe):
            for j in range(3):
                f.write(int.to_bytes(BYTESIZE*natom, BYTESIZE, "little"))
                f.write(np.float32(arr[i, :, j]).tobytes())
                f.write(int.to_bytes(BYTESIZE*natom, BYTESIZE, "little"))
    print("\t Done \n")

def existsCommand(name):
    from shutil import which
    return which(name) is not None



# **************************************************************************
# *
# * Authors:    Mohamad Harastani            (mohamad.harastani@igbmc.fr)
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
from pwem.protocols import ProtAnalysis3D
from xmipp3.convert import writeSetOfVolumes, xmippToLocation
from pwem.objects import Volume
import pwem.emlib.metadata as md
import pyworkflow.protocol.params as params
from pwem.utils import runProgram
from pwem import Domain
from .convert import eulerAngles2matrix, matrix2eulerAngles
import numpy as np
import multiprocessing

WEDGE_MASK_NONE = 0
WEDGE_MASK_THRE = 1

REFERENCE_NONE = 0
REFERENCE_EXISTS = 1
REFERENCE_IMPORTED = 2

PERFORM_STA = 0
COPY_STA = 1

IMPORT_XMIPP_MD = 0
IMPORT_DYNAMO_TBL = 1
IMPORT_TOMBOX_MTV = 2

class FlexProtSubtomogramAveraging(ProtAnalysis3D):
    """ Protocol for subtomogram averaging. This protocol has two modes of operation.
     the first is to perform subtomogram averaging using Fast Rotational Matching.
     The second mode is to import a previously performed alignment using this protocol, Dynamo, or Artiatomi.
     If an alignment is imported, the rigid-body parameters will be used to re-create the average structure.
     """

    _label = 'subtomogram averaging'

    # --------------------------- DEFINE param functions --------------------------------------------
    def _defineParams(self, form):
        form.addSection(label='')
        group = form.addGroup('Settings')
        # form.addSection(label='Input')
        group.addParam('inputVolumes', params.PointerParam,
                      pointerClass='SetOfVolumes',
                      label="Input volume(s)", important=True,
                      help='Select volumes')
        group.addParam('StA_choice', params.EnumParam,
                      choices=['Perform StA using Fast Rotational Matching (FRM)',
                               'Import parameters of a previously performed StA'],
                      default=PERFORM_STA,
                      label='Choose what processes you want to perform:', display=params.EnumParam.DISPLAY_COMBO,
                       help='If you choose to "Perform StA" using FRM you have to set the parameters in the last tab.'
                            ' Otherwise, if you import a pervious StA, the remaining parameters are not used')
        group = form.addGroup('Importing a previous StA protocol',
                              condition='StA_choice==%d'% COPY_STA)
        group.addParam('import_choice', params.EnumParam,
                       label='From which software?',
                       choices=['Import Scipion/Xmipp metadata',
                                'Import Dynamo table',
                                'Import TOM-ToolBox motive list'],
                       default=IMPORT_XMIPP_MD,
                       help='You have to provide a pervious table of rigid-body alignment parameters in one of the list'
                            'of supported formats. The software will evaluate the average based on the provided file. '
                            'If the average is correct (corresponds to what you have before) then the alignment went correctly,'
                            ' and in this case you can proceed in further processing (refinement and heterogeneity analysis).'
                       )
        group.addParam('xmippMD', params.PathParam, allowsNull=True,
                       condition='import_choice==%d' %IMPORT_XMIPP_MD,
                      label='Import Scipion/Xmipp metadata file',
                      help='import a the metadata file that contains the StA parameters. This option will evaluate '
                           'the average and allows you to perform post-StA processes (refinement and heterogeneity analysis).')
        group.addParam('dynamoTable', params.PathParam, allowsNull=True,
                       condition='import_choice==%d' % IMPORT_DYNAMO_TBL,
                       label='Import a Dynamo table [Beta]',
                      help='import a Dynamo table that contains the StA parameters. This option will evaluate '
                           'the average and transform the Dynamo table to Scipion metadata format. and allows you to '
                           'perform post-StA processes (refinement and heterogeneity analysis).')
        group.addParam('tomBoxTable', params.PathParam, allowsNull=True,
                       condition='import_choice==%d' % IMPORT_TOMBOX_MTV,
                       label='Import a TOM-toolbox table (motive list) [Beta]',
                      help='import a TOM-toolbox table that contains the STA parameters. This option will evaluate '
                           'the average and transform the motive list to Scipion metadata format. and allows you to '
                           'perform post-StA processes (refinement and heterogeneity analysis).')
        group = form.addGroup('Subtomogram Averaging using Fast Rotational Matching (FRM)',
                              condition='StA_choice==%d'% PERFORM_STA)
        group.addParam('StartingReference', params.EnumParam,
                      choices=['start from scratch', 'browse for a volume file and use it as reference',
                               'select a volume from the workspace and use as reference'],
                      default=REFERENCE_NONE,
                      label='Starting reference', display=params.EnumParam.DISPLAY_COMBO,
                      help='Align from scratch of choose a template')
        group.addParam('ReferenceVolume', params.FileParam,
                      pointerClass='params.FileParam', allowsNull=True,
                      condition='StartingReference==%d' % REFERENCE_EXISTS,
                      label="starting reference file",
                      help='Choose a starting reference from an external volume file')
        group.addParam('ReferenceImported', params.PointerParam,
                      pointerClass='SetOfVolumes,Volume', allowsNull=True,
                      condition='StartingReference==%d' % REFERENCE_IMPORTED,
                      label="selected starting reference",
                      help='Choose an imported volume as a starting reference')
        group.addParam('applyMask', params.BooleanParam, label='Use a mask?', default=False,
                       help='A mask that can be applied on the reference without cropping it. The same mask will be'
                            ' applied on the aligned subtomograms at each iteration (do not apply this mask in advance)'
                       )
        group.addParam('Mask', params.PointerParam,
                       condition='applyMask',
                       pointerClass='Volume', allowsNull=True,
                       label="Select mask")
        group.addParam('NumOfIters', params.IntParam, default=10,
                      label='Number of iterations', help='How many times you want to iterate while performing'
                                                         ' subtomogram alignment and averaging.')
        group.addParam('WedgeMode', params.EnumParam,
                      choices=['Do not compensate', 'Compensate'],
                      default=WEDGE_MASK_THRE,
                      label='Missing-wedge Compensation', display=params.EnumParam.DISPLAY_COMBO,
                      help='Choose to compensate for the missing wedge if aligning subtomograms.'
                           ' However, if you are working with previously aligned subtomograms, then its better not to.')
        line = group.addLine('Low and high tilt values:', condition='WedgeMode==%d' % WEDGE_MASK_THRE,
                             help='The lower and upper tilt angles used in obtaining the tilt series')
        line.addParam('tiltLow', params.IntParam, default=-60,
                      label='Lower tilt value')
        line.addParam('tiltHigh', params.IntParam, default=60,
                      label='Upper tilt value')
        line = group.addLine('FRM parameters:',
                             help='The normalized frequency should be between 0 and 0.5 '
                                  'The more it is, the bigger the search frequency is, the more time it demands, '
                                  'keeping it as default is recommended. The maximum shift is a number between 1 and half the '
                                  'size of your volume. It represents the maximum distance searched in x,y and'
                                  ' z directions.'
                             )
        line.addParam('frm_freq', params.FloatParam, default=0.25,
                      label='Maximum searched frequency (0->0.5)')
        line.addParam('frm_maxshift', params.IntParam, default=10,
                      label='Maximum shift search (in pixels)',
                      help='')
        form.addParallelSection(threads=0, mpi=multiprocessing.cpu_count()//2-1)

    # --------------------------- INSERT steps functions --------------------------------------------

    def _insertAllSteps(self):
        # Define some outputs filenames
        self.imgsFn = self._getExtraPath('volumes.xmd')
        self.outputVolume = self._getExtraPath('final_average.mrc')
        self.outputMD = self._getExtraPath('final_md.xmd')

        self._insertFunctionStep('convertInputStep')
        if self.StA_choice.get() == PERFORM_STA:
            self._insertFunctionStep('doAlignmentStep')
        elif self.StA_choice.get() == COPY_STA and self.import_choice.get() == IMPORT_DYNAMO_TBL:
            self._insertFunctionStep('adaptDynamoStep', self.dynamoTable.get())
        elif self.StA_choice.get() == COPY_STA and self.import_choice.get() == IMPORT_TOMBOX_MTV:
            self._insertFunctionStep('adaptTomboxStep', self.tomBoxTable.get())
        else:
            self._insertFunctionStep('adaptXmippStep', self.xmippMD.get())
        self._insertFunctionStep('createOutputStep')

    # --------------------------- STEPS functions --------------------------------------------
    def convertInputStep(self):
        # Write a metadata with the volumes to align
        writeSetOfVolumes(self.inputVolumes.get(), self.imgsFn)

    def doAlignmentStep(self):
        tempdir = self._getTmpPath()
        imgFn = self.imgsFn
        frm_freq = self.frm_freq.get()
        frm_maxshift = self.frm_maxshift.get()
        max_itr = self.NumOfIters.get()
        iter_result = self._getExtraPath('result.xmd')
        reference = None
        if self.StartingReference == REFERENCE_EXISTS:
            reference = self.ReferenceVolume.get()
        if self.StartingReference == REFERENCE_IMPORTED:
            reference = self.ReferenceImported.get().getFileName()

        print('tempdir is ', tempdir)
        print('imgFn is ', imgFn)
        print('frm_freq is ', frm_freq)
        print('frm_maxshift is ', frm_maxshift)
        print('max_itr is ', max_itr)
        print('iter_result is ', iter_result)
        print('reference is ', reference)

        # if the reference is None, then we got to make an initial reference:
        if reference is None:
            initialref = self._getExtraPath('initialref.mrc')
            mdImgs = md.MetaData(imgFn)
            counter = 0
            first = True
            tempVol = self._getExtraPath('temp.mrc')
            for objId in mdImgs:
                counter = counter + 1
                imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
                if counter == 1:
                    args = '-i %(imgPath)s -o %(tempVol)s --type vol' % locals()
                    runProgram('xmipp_image_convert',args)
                else:
                    params = '-i %(imgPath)s --plus %(tempVol)s -o %(tempVol)s ' % locals()
                    runProgram('xmipp_image_operate', params)
            params = '-i %(tempVol)s --divide %(counter)s -o %(initialref)s ' % locals()
            runProgram('xmipp_image_operate', params)
            os.system("rm -f %(tempVol)s" % locals())
            reference = initialref

        for i in range(1, max_itr + 1):
            arg = 'params_itr_' + str(i) + '.xmd'
            md_itr = self._getExtraPath(arg)
            arg = 'average_itr_' + str(i) + '.mrc'
            avr_itr = self._getExtraPath(arg)
            args = "-i %(imgFn)s -o %(md_itr)s --odir %(tempdir)s --resume --ref %(reference)s" \
                   " --frm_parameters %(frm_freq)f %(frm_maxshift)d "

            if self.WedgeMode == WEDGE_MASK_THRE:
                tilt0 = self.tiltLow.get()
                tiltF = self.tiltHigh.get()
                # args += " %(tilt0)d %(tiltF)d "
                args += "--tilt_values %(tilt0)d %(tiltF)d "

            if self.applyMask.get():
                args += "--mask " + self.Mask.get().getFileName()

            self.runJob("xmipp_volumeset_align", args % locals(),
                        env = Domain.importFromPlugin('xmipp3').Plugin.getEnviron())

            # By now, the alignment is done, the averaging should take place
            # However, if the alignemnt has missing wedge compensation, we shall update the metadata:
            if self.WedgeMode == WEDGE_MASK_THRE:
                mdImgs = md.MetaData(md_itr)
                for objId in mdImgs:
                    rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
                    tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
                    psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)
                    x = mdImgs.getValue(md.MDL_SHIFT_X, objId)
                    y = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
                    z = mdImgs.getValue(md.MDL_SHIFT_Z, objId)
                    T = eulerAngles2matrix(rot, tilt, psi, x, y, z)
                    # Rotate 90 degrees (compensation for missing wedge)
                    T0 = eulerAngles2matrix(0, 90, 0, 0, 0, 0)
                    T = np.linalg.inv(np.matmul(T, T0))
                    rot, tilt, psi, x, y, z = matrix2eulerAngles(T)
                    mdImgs.setValue(md.MDL_ANGLE_ROT, rot, objId)
                    mdImgs.setValue(md.MDL_ANGLE_TILT, tilt, objId)
                    mdImgs.setValue(md.MDL_ANGLE_PSI, psi, objId)
                    mdImgs.setValue(md.MDL_SHIFT_X, x, objId)
                    mdImgs.setValue(md.MDL_SHIFT_Y, y, objId)
                    mdImgs.setValue(md.MDL_SHIFT_Z, z, objId)
                    mdImgs.setValue(md.MDL_ANGLE_Y, 0.0, objId)
                mdImgs.write(md_itr)

            mdImgs = md.MetaData(md_itr)
            counter = 0

            for objId in mdImgs:
                counter = counter + 1

                imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
                rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
                tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
                psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)

                x_shift = mdImgs.getValue(md.MDL_SHIFT_X, objId)
                y_shift = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
                z_shift = mdImgs.getValue(md.MDL_SHIFT_Z, objId)

                tempVol = self._getExtraPath('temp.mrc')
                extra = self._getExtraPath()

                params = '-i %(imgPath)s -o %(tempVol)s --inverse --rotate_volume euler %(rot)s %(tilt)s %(psi)s' \
                         ' --shift %(x_shift)s %(y_shift)s %(z_shift)s -v 0' % locals()
                runProgram('xmipp_transform_geometry', params)

                if counter == 1:
                    os.system("cp %(tempVol)s %(avr_itr)s" % locals())

                else:
                    params = '-i %(tempVol)s --plus %(avr_itr)s -o %(avr_itr)s ' % locals()
                    runProgram('xmipp_image_operate', params)

            params = '-i %(avr_itr)s --divide %(counter)s -o %(avr_itr)s ' % locals()
            runProgram('xmipp_image_operate', params)
            os.system("rm -f %(tempVol)s" % locals())
            # Updating the reference then realigning:
            reference = avr_itr

        outputVolume = self.outputVolume
        outputMD = self.outputMD
        os.system("cp %(avr_itr)s %(outputVolume)s " % locals())
        os.system("cp %(md_itr)s %(outputMD)s " % locals())
        # Averaging is done
        inputSet = md.MetaData(self.imgsFn)
        mdImgs = md.MetaData(self.outputMD)
        # setting item_id (lost due to mpi usually)
        for objId in mdImgs:
            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            index, fn = xmippToLocation(imgPath)
            # Conside the index is the id in the input set
            for objId2 in inputSet:
                NewImgPath = inputSet.getValue(md.MDL_IMAGE, objId2)
                if (NewImgPath == imgPath):
                    target_ID = inputSet.getValue(md.MDL_ITEM_ID, objId2)
                    break
            mdImgs.setValue(md.MDL_ITEM_ID, target_ID, objId)
        mdImgs.sort(md.MDL_ITEM_ID)
        mdImgs.write(self.outputMD)


    def adaptDynamoStep(self, dynamoTable):
        volumes_in = self.imgsFn
        volume_out = self.outputVolume
        md_out = self.outputMD
        from continuousflex.protocols.utilities.dynamo import tbl2metadata
        tbl2metadata(dynamoTable, volumes_in, md_out)

        mdImgs = md.MetaData(md_out)
        counter = 0
        first = True

        for objId in mdImgs:
            counter = counter + 1

            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
            tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
            psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)
            x_shift = mdImgs.getValue(md.MDL_SHIFT_X, objId)
            y_shift = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
            z_shift = mdImgs.getValue(md.MDL_SHIFT_Z, objId)

            tempVol = self._getExtraPath('temp.mrc')
            extra = self._getExtraPath()


            if first:
                print("Averaging based on Dynamo parameters")
                first = False

            params = '-i %(imgPath)s -o %(tempVol)s --inverse --rotate_volume euler %(rot)s %(tilt)s %(psi)s' \
                     ' --shift %(x_shift)s %(y_shift)s %(z_shift)s -v 0' % locals()

            runProgram('xmipp_transform_geometry', params)

            if counter == 1:
                os.system("cp %(tempVol)s %(volume_out)s" % locals())

            else:
                params = '-i %(tempVol)s --plus %(volume_out)s -o %(volume_out)s ' % locals()
                runProgram('xmipp_image_operate', params)

        params = '-i %(volume_out)s --divide %(counter)s -o %(volume_out)s ' % locals()
        runProgram('xmipp_image_operate', params)
        os.system("rm -f %(tempVol)s" % locals())


    def adaptTomboxStep(self, Table):
        volumes_in = self.imgsFn
        volume_out = self.outputVolume
        md_out = self.outputMD
        from continuousflex.protocols.utilities.tombox import motivelist2metadata
        motivelist2metadata(Table, volumes_in, md_out)

        # Averaging based on the metadata:
        mdImgs = md.MetaData(md_out)
        counter = 0

        for objId in mdImgs:
            counter = counter + 1

            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
            tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
            psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)

            x_shift = mdImgs.getValue(md.MDL_SHIFT_X, objId)
            y_shift = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
            z_shift = mdImgs.getValue(md.MDL_SHIFT_Z, objId)

            tempVol = self._getExtraPath('temp.mrc')
            extra = self._getExtraPath()

            params = '-i %(imgPath)s -o %(tempVol)s --inverse --rotate_volume euler %(rot)s %(tilt)s %(psi)s' \
                     ' --shift %(x_shift)s %(y_shift)s %(z_shift)s' % locals()

            runProgram('xmipp_transform_geometry', params)

            if counter == 1:
                os.system("cp %(tempVol)s %(volume_out)s" % locals())

            else:
                params = '-i %(tempVol)s --plus %(volume_out)s -o %(volume_out)s ' % locals()
                runProgram('xmipp_image_operate', params)

        params = '-i %(volume_out)s --divide %(counter)s -o %(volume_out)s ' % locals()
        runProgram('xmipp_image_operate', params)
        os.system("rm -f %(tempVol)s" % locals())
         # Averaging is done
        pass


    def adaptXmippStep(self, Table):
        volumes_in = self.imgsFn
        volume_out = self.outputVolume
        md_out = Table

        # Averaging based on the metadata:
        mdImgs = md.MetaData(md_out)

        # if the volumes were aligned with angle_y=90 degrees, then rotate by 90 and inverse, then set angle y to 0
        flag = None
        try:
            flag = mdImgs.getValue(md.MDL_ANGLE_Y, 1)
        except:
            pass

        if flag == 90:
            mdImgs = md.MetaData(self.imgsFn)
            for objId in mdImgs:
                rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
                tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
                psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)
                x = mdImgs.getValue(md.MDL_SHIFT_X, objId)
                y = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
                z = mdImgs.getValue(md.MDL_SHIFT_Z, objId)
                T = eulerAngles2matrix(rot, tilt, psi, x, y, z)
                # Rotate 90 degrees (compensation for missing wedge)
                T0 = eulerAngles2matrix(0, 90, 0, 0, 0, 0)
                T = np.linalg.inv(np.matmul(T, T0))
                rot, tilt, psi, x, y, z = matrix2eulerAngles(T)
                mdImgs.setValue(md.MDL_ANGLE_ROT, rot, objId)
                mdImgs.setValue(md.MDL_ANGLE_TILT, tilt, objId)
                mdImgs.setValue(md.MDL_ANGLE_PSI, psi, objId)
                mdImgs.setValue(md.MDL_SHIFT_X, x, objId)
                mdImgs.setValue(md.MDL_SHIFT_Y, y, objId)
                mdImgs.setValue(md.MDL_SHIFT_Z, z, objId)
                mdImgs.setValue(md.MDL_ANGLE_Y, 0.0, objId)

        mdImgs.write(self._getExtraPath('final_md.xmd'))
        counter = 0

        for objId in mdImgs:
            counter = counter + 1

            imgPath = mdImgs.getValue(md.MDL_IMAGE, objId)
            rot = mdImgs.getValue(md.MDL_ANGLE_ROT, objId)
            tilt = mdImgs.getValue(md.MDL_ANGLE_TILT, objId)
            psi = mdImgs.getValue(md.MDL_ANGLE_PSI, objId)

            x_shift = mdImgs.getValue(md.MDL_SHIFT_X, objId)
            y_shift = mdImgs.getValue(md.MDL_SHIFT_Y, objId)
            z_shift = mdImgs.getValue(md.MDL_SHIFT_Z, objId)

            tempVol = self._getExtraPath('temp.mrc')
            extra = self._getExtraPath()

            params = '-i %(imgPath)s -o %(tempVol)s --inverse --rotate_volume euler %(rot)s %(tilt)s %(psi)s' \
                 ' --shift %(x_shift)s %(y_shift)s %(z_shift)s' % locals()

            runProgram('xmipp_transform_geometry', params)

            if counter == 1:
                os.system("cp %(tempVol)s %(volume_out)s" % locals())

            else:
                params = '-i %(tempVol)s --plus %(volume_out)s -o %(volume_out)s ' % locals()
                runProgram('xmipp_image_operate', params)

        params = '-i %(volume_out)s --divide %(counter)s -o %(volume_out)s ' % locals()
        runProgram('xmipp_image_operate', params)
        os.system("rm -f %(tempVol)s" % locals())


    def createOutputStep(self):
        inputSet = self.inputVolumes.get()
        outvolume = Volume()
        outvolume.setSamplingRate(inputSet.getSamplingRate())
        outvolume.setFileName(self.outputVolume)
        self._defineOutputs(SubtomogramAverage=outvolume)


    # --------------------------- INFO functions --------------------------------------------
    def _summary(self):
        summary = []
        return summary

    def _citations(self):
        return ['harastani2022continuousflex','CHEN2013235']

    def _methods(self):
        pass



# **************************************************************************
# * Authors: Rémi Vuillemot             (remi.vuillemot@upmc.fr)
# *
# * IMPMC, UPMC Sorbonne University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# **************************************************************************


from pyworkflow.viewer import (ProtocolViewer, DESKTOP_TKINTER, WEB_DJANGO)
import pyworkflow.protocol.params as params
from continuousflex.protocols.protocol_mdspace import FlexProtMDSPACE
from continuousflex.viewers.viewer_genesis import FlexGenesisViewer
from continuousflex.protocols.utilities.genesis_utilities import *

from .plotter import FlexPlotter
from pwem.viewers import VmdView, ChimeraView
from pyworkflow.utils import getListFromRangeString
import numpy as np
import os
import glob
import pwem.emlib.metadata as md
import re

from matplotlib.pyplot import cm

class FlexMDSPACEViewer(FlexGenesisViewer):
    """ Visualization of results from the MDSPACE protocol
    """
    _label = 'MDSPACE Viewer'
    _targets = [FlexProtMDSPACE]
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]

    def _defineParams(self, form):
        FlexGenesisViewer._defineParams(self, form)
        group = form.addGroup('MDSPACE')
        group.addParam('displayPCA', params.LabelParam,
                      label='Display PCA space')
        group.addParam('pcaAxes', params.StringParam, default="1 2",
                       label='Axes to display' )
        group.addParam('displayFE', params.LabelParam,
                      label='Display free energy')
        group.addParam('feAxes', params.StringParam, default="1 2",
                       label='Axes to display' )
        group.addParam('freeEnergySize', params.IntParam, default=20,
                       label='Sampling size' )
    def _getVisualizeDict(self):
        dict = FlexGenesisViewer._getVisualizeDict(self)
        dict['displayPCA'] = self._plotPCA
        dict['displayFE'] = self._plotFE
        return dict

    def _plotPCA(self, p):
        axes_str = str.split(self.pcaAxes.get())
        axes = []
        for i in axes_str: axes.append(int(i.strip())-1)

        plotter = FlexPlotter(1,self.protocol.numberOfIter.get()
                              , figsize=(self.protocol.numberOfIter.get()*3,3))
        for i in range(self.protocol.numberOfIter.get()):
            self.protocol._iter= i
            pca = np.loadtxt(self.protocol.getPCAPrefix() + "_matrix.txt")[:,axes]
            ax = plotter.createSubPlot("PCA iter "+str(i+1), "component " + axes_str[0],
                                   "component " + axes_str[1], xpos=1, ypos=i+1)
            ax.scatter(pca[:,0],pca[:,1])
        plotter.show()
    def _plotFE(self, p):
        axes_str = str.split(self.feAxes.get())
        axes = []
        for i in axes_str: axes.append(int(i.strip())-1)
        size =self.freeEnergySize.get()

        plotter = FlexPlotter(1,self.protocol.numberOfIter.get()
                              , figsize=(self.protocol.numberOfIter.get()*3,3))
        for i in range(self.protocol.numberOfIter.get()):
            self.protocol._iter= i

            data = np.loadtxt(self.protocol.getPCAPrefix() + "_matrix.txt")[:,axes]
            xmin = np.min(data[:,0])
            xmax = np.max(data[:,0])
            ymin = np.min(data[:,1])
            ymax = np.max(data[:,1])
            x = np.linspace(xmin, xmax, size)
            y = np.linspace(ymin, ymax, size)
            count = np.zeros((size, size))
            for j in range(data.shape[0]):
                count[np.argmin(np.abs(x.T - data[j, 0])),
                      np.argmin(np.abs(y.T - data[j, 1]))] += 1
            img = -np.log(count / count.max())
            img[img == np.inf] = img[img != np.inf].max()

            xx, yy = np.mgrid[xmin:xmax:size * 1j, ymin:ymax:size * 1j]

            ax = plotter.createSubPlot("Free energy iter "+str(i+1), "component " + axes_str[0],
                                   "component " + axes_str[1], xpos=1, ypos=i+1)
            im = ax.contourf(xx, yy, img, cmap='jet')
            # im = ax.imshow(img.T[::-1, :],
            #                cmap="jet", interpolation="bicubic",
            #                extent=[xmin, xmax, ymin, ymax])
            cbar = plotter.figure.colorbar(im)
            cbar.set_label("$\Delta G / k_{B}T$")
        plotter.show()